/*
 * Jacobin VM - A Java virtual machine
 * Copyright (c) 2023 by the Jacobin authors. All rights reserved.
 * Licensed under Mozilla Public License 2.0 (MPL 2.0)
 */

package wholeClassTests

import (
    "fmt"
    "io"
    "log"
    "os"
    "os/exec"
    "strings"
    "testing"
)

// Test for testBitShift class, which tests bit shifting. Source code:
//
//	public class testBitShifts{
//		public static void main( String[] args) {
//			int a = -100;
//			int b = a >> 2;
//			System.out.print("-100 >> 2: "); System.out.println(b);
//
//			int c = a << 3;
//			System.out.print("-100 << 3: "); System.out.println(c);
//
//			int pos = 100;
//			int d = pos >> 2;
//			System.out.print("+100 >> 2: "); System.out.println(d);
//
//			int e = pos << 3;
//			System.out.print("+100 << 3: "); System.out.println(e);
//		}
//	}
//
// To run your class, enter its name in _TESTCLASS, any args in their respective variables and then run the tests.
// This test harness expects that environmental variable JACOBIN_EXE gives the full name and path of the executable
// we're running the tests on. The folder which contains the test class should be specified in the environmental
// variable JACOBIN_TESTDATA (without a terminating slash).
func initVarsTestBitShifts() error {
    if testing.Short() { // don't run if running quick tests only. (Used primarily so GitHub doesn't run and bork)
        return fmt.Errorf("test not run due to -short")
    }

    _JACOBIN = os.Getenv("JACOBIN_EXE") // returns "" if JACOBIN_EXE has not been specified.
    _JVM_ARGS = ""
    _TESTCLASS = "testBitShifts.class" // the class to test
    _APP_ARGS = ""

    if _JACOBIN == "" {
        return fmt.Errorf("test failure due to missing Jacobin executable. Please specify it in JACOBIN_EXE")
    } else if _, err := os.Stat(_JACOBIN); err != nil {
        return fmt.Errorf("missing Jacobin executable, which was specified as %s", _JACOBIN)
    }

    if _TESTCLASS != "" {
        testClass := os.Getenv("JACOBIN_TESTDATA") + string(os.PathSeparator) + _TESTCLASS
        if _, err := os.Stat(testClass); err != nil {
            return fmt.Errorf("missing class to test, which was specified as %s", testClass)
        } else {
            _TESTCLASS = testClass
        }
    }
    return nil
}

func TestRunBitShifts(t *testing.T) {
    if testing.Short() { // don't run if running quick tests only. (Used primarily so GitHub doesn't run and bork)
        t.Skip()
    }

    initErr := initVarsTestBitShifts()
    if initErr != nil {
        t.Fatalf("Test failure due to: %s", initErr.Error())
    }
    var cmd *exec.Cmd

    if testing.Short() { // don't run if running quick tests only. (Used primarily so GitHub doesn't run and bork)
        t.Skip()
    }

    // test that executable exists
    if _, err := os.Stat(_JACOBIN); err != nil {
        t.Errorf("Missing Jacobin executable, which was specified as %s", _JACOBIN)
    }

    // run the various combinations of args. This is necessary b/c the empty string is viewed as
    // an actual specified option on the command line.
    if len(_JVM_ARGS) > 0 {
        if len(_APP_ARGS) > 0 {
            cmd = exec.Command(_JACOBIN, _JVM_ARGS, _TESTCLASS, _APP_ARGS)
        } else {
            cmd = exec.Command(_JACOBIN, _JVM_ARGS, _TESTCLASS)
        }
    } else {
        if len(_APP_ARGS) > 0 {
            cmd = exec.Command(_JACOBIN, _TESTCLASS, _APP_ARGS)
        } else {
            cmd = exec.Command(_JACOBIN, _TESTCLASS)
        }
    }

    // get the stdout and stderr contents from the file execution
    stderr, err := cmd.StderrPipe()
    stdout, err := cmd.StdoutPipe()
    if err != nil {
        log.Fatal(err)
    }

    // run the command
    if err = cmd.Start(); err != nil {
        t.Errorf("Got error running Jacobin: %s", err.Error())
    }

    // Here begin the actual tests on the output to stderr and stdout
    slurp, _ := io.ReadAll(stderr)
    slurpErr := string(slurp)
    if len(slurp) != 0 {
        t.Errorf("Got unexpected output to stderr: %s", slurpErr)
    }

    slurp, _ = io.ReadAll(stdout)
    slurpOut := string(slurp)

    if !strings.Contains(slurpOut, "-100 >> 2: -25") &&
        !strings.Contains(string(slurp), "+100 << 3: 800") {
        t.Errorf("Did not get expected output to stdout. Got: %s", slurpOut)
    }
}
